/*
 * kgraft_patch_bsc1003253
 *
 * Fix for CVE-2016-7117, bsc#1003253
 *
 *  Upstream commit:
 *  34b88a68f26a ("net: Fix use after free in the recvmmsg exit path")
 *
 *  SLE12 commit:
 *  bf3c5c7200f38facfc633d1908e622e9168fa52d
 *
 *  Copyright (c) 2016 SUSE
 *  Author: Miroslav Benes <mbenes@suse.cz>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kallsyms.h>
#include <linux/socket.h>
#include <linux/file.h>
#include <net/sock.h>
#include <net/compat.h>

static int (*kgr_poll_select_set_timeout)(struct timespec *to, long sec, long nsec);
static struct socket *(*kgr_sockfd_lookup_light)(int fd, int *err, int *fput_needed);
static int (*kgr___sys_recvmsg)(struct socket *sock, struct msghdr __user *msg, struct msghdr *msg_sys, unsigned int flags, int nosec);

static struct {
	char *name;
	void **addr;
} kgr_funcs[] = {
	{ "poll_select_set_timeout", (void *)&kgr_poll_select_set_timeout },
	{ "sockfd_lookup_light", (void *)&kgr_sockfd_lookup_light },
	{ "___sys_recvmsg", (void *)&kgr___sys_recvmsg },
};

/* patched function */
int kgr__sys_recvmmsg(int fd, struct mmsghdr __user *mmsg, unsigned int vlen,
		   unsigned int flags, struct timespec *timeout)
{
	int fput_needed, err, datagrams;
	struct socket *sock;
	struct mmsghdr __user *entry;
	struct compat_mmsghdr __user *compat_entry;
	struct msghdr msg_sys;
	struct timespec end_time;

	if (timeout &&
	    kgr_poll_select_set_timeout(&end_time, timeout->tv_sec,
				    timeout->tv_nsec))
		return -EINVAL;

	datagrams = 0;

	sock = kgr_sockfd_lookup_light(fd, &err, &fput_needed);
	if (!sock)
		return err;

	err = sock_error(sock->sk);
	if (err)
		goto out_put;

	entry = mmsg;
	compat_entry = (struct compat_mmsghdr __user *)mmsg;

	while (datagrams < vlen) {
		/*
		 * No need to ask LSM for more than the first datagram.
		 */
		if (MSG_CMSG_COMPAT & flags) {
			err = kgr___sys_recvmsg(sock, (struct msghdr __user *)compat_entry,
					     &msg_sys, flags & ~MSG_WAITFORONE,
					     datagrams);
			if (err < 0)
				break;
			err = __put_user(err, &compat_entry->msg_len);
			++compat_entry;
		} else {
			err = kgr___sys_recvmsg(sock,
					     (struct msghdr __user *)entry,
					     &msg_sys, flags & ~MSG_WAITFORONE,
					     datagrams);
			if (err < 0)
				break;
			err = put_user(err, &entry->msg_len);
			++entry;
		}

		if (err)
			break;
		++datagrams;

		/* MSG_WAITFORONE turns on MSG_DONTWAIT after one packet */
		if (flags & MSG_WAITFORONE)
			flags |= MSG_DONTWAIT;

		if (timeout) {
			ktime_get_ts(timeout);
			*timeout = timespec_sub(end_time, *timeout);
			if (timeout->tv_sec < 0) {
				timeout->tv_sec = timeout->tv_nsec = 0;
				break;
			}

			/* Timeout, return less than vlen datagrams */
			if (timeout->tv_nsec == 0 && timeout->tv_sec == 0)
				break;
		}

		/* Out of band data, return right away */
		if (msg_sys.msg_flags & MSG_OOB)
			break;
	}

	/*
	 * Fix CVE-2016-7117
	 *  completely new error handling
	 * out_put:
	 *  fput_light(sock->file, fput_needed);
	 *
	 * if (err == 0)
	 *	return datagrams;
	 *
	 *	if (datagrams != 0) {
	 *		 *
	 *		 * We may return less entries than requested (vlen) if the
	 *		 * sock is non block and there aren't enough datagrams...
	 *		 *
	 *		if (err != -EAGAIN) {
	 *			 *
	 *			 * ... or  if recvmsg returns an error after we
	 *			 * received some datagrams, where we record the
	 *			 * error to return on the next call or if the
	 *			 * app asks about it using getsockopt(SO_ERROR).
	 *			 *
	 *			sock->sk->sk_err = -err;
	 *		}
	 *
	 *		return datagrams;
	 *	}
	 *
	 *	return err;
	 */
	if (err == 0)
		goto out_put;

	if (datagrams == 0) {
		datagrams = err;
		goto out_put;
	}

	/*
	 * We may return less entries than requested (vlen) if the
	 * sock is non block and there aren't enough datagrams...
	 */
	if (err != -EAGAIN) {
		/*
		 * ... or  if recvmsg returns an error after we
		 * received some datagrams, where we record the
		 * error to return on the next call or if the
		 * app asks about it using getsockopt(SO_ERROR).
		 */
		sock->sk->sk_err = -err;
	}
out_put:
	fput_light(sock->file, fput_needed);

	return datagrams;
}

static int kgr_patch_bsc1003253_kallsyms(void)
{
	unsigned long addr;
	int i;

	for (i = 0; i < ARRAY_SIZE(kgr_funcs); i++) {
		addr = kallsyms_lookup_name(kgr_funcs[i].name);
		if (!addr) {
			pr_err("kgraft-patch: symbol %s not resolved\n",
				kgr_funcs[i].name);
			return -ENOENT;
		}

		*(kgr_funcs[i].addr) = (void *)addr;
	}

	return 0;
}

int kgr_patch_bsc1003253_init(void)
{
	return kgr_patch_bsc1003253_kallsyms();
}
